#!-*-coding:utf-8-*-

from abc import ABC, abstractmethod
from osgeo.osr import SpatialReference
from osgeo.ogr import Layer


class ISRS(ABC):

    def _srs_from_layer(self, lyr: Layer, ds_name: str) -> SpatialReference:
        in_spatial_ref = lyr.GetSpatialRef()

        if in_spatial_ref is None:
            raise DataSourceWithoutSpatialRef(
                f'The datasource {ds_name} not have spatial'
                ' reference. Plese, use the input SRID.'
            )
        return in_spatial_ref

    @abstractmethod
    def input_srs(self) -> SpatialReference:
        pass

    @abstractmethod
    def output_srs(self) -> SpatialReference:
        pass



# -*-coding:utf-8-*-

from abc import ABC, abstractmethod
from typing import Generator
from osgeo.ogr import DataSource


class IVectorComposite(ABC):

    @abstractmethod
    def items(self, input_data: str) -> Generator[str, None, None]:
        """Docs"""
        pass

    @abstractmethod
    def collection(self, input_data: str) -> str:
        """ Docs """
        pass

    @abstractmethod
    def write(self, input_data: str, out_path: str) -> str:
        """ Docs """
        pass



#!-*-coding:utf-8-*-


class CharDecodeError(Exception):
    pass


class FileNotFound(Exception):
    pass


class GeojsonInvalid(Exception):
    pass


class WKTInvalid(Exception):
    pass


class GeometryTypeInvalid(Exception):
    pass


class DataSourceWithoutSpatialRef(Exception):
    pass


class TypeFromFileIsNotSupported(Exception):
    pass


class ShapefileInvalid(Exception):
    pass


class ShapefileIsEmpty(Exception):
    pass


class DirectoryDoesNotExists(Exception):
    pass


class ImpossibleCreateShapefileFromGeometryCollection(Exception):
    pass


class ErrorOnIntersection(Exception):
    pass


class ExistsManyGeometriesTypes(Exception):
    pass


class DataSourceNotIntersectsWithAnyUTMZones(Exception):
    pass


class SRIDNotFound(Exception):
    pass


class WKTPRJNotFound(Exception):
    pass


#!-*-coding:utf-8-*-

from vectorio.gdal.error.custom_error import GDALCustomError
from vectorio.gdal.exceptions import (
    GDALSelfIntersectionGeometry, GDALUnknownException, GDALBadClosedPolygon
)
from osgeo import gdal


KEY_SELF_INTERSECTION = 'Self-intersection'
MSG_BAD_CLOSED_POLYGON = 'IllegalArgumentException: Points of LinearRing do not form a closed linestring'


def gdal_warning_as_exception(inner):

    def func(*args, **kwargs):
        error = GDALCustomError()
        handler = error.handler
        gdal.PushErrorHandler(handler)
        gdal.UseExceptions()

        result = inner(*args, **kwargs)

        if error.level == gdal.CE_None:  # without error
            return result
        if MSG_BAD_CLOSED_POLYGON == error.msg:
            raise GDALBadClosedPolygon(error.msg)
        elif KEY_SELF_INTERSECTION in error.msg:
            raise GDALSelfIntersectionGeometry(error.msg)
        else:
            raise GDALUnknownException(error.msg)

    return func



#!-*-coding:utf-8-*-

from vectorio.vector.srs.isrs import ISRS
from osgeo.osr import SpatialReference
from osgeo.ogr import Layer
from vectorio.vector.exceptions import SRIDNotFound
from osgeo import osr


class SRSFromSRID(ISRS):

    def __init__(self, ds_name: str, in_lyr: Layer, out_srid: int, in_srid=None):
        self._out_srid = out_srid
        self._in_srid = in_srid
        self._ds_name = ds_name
        self._in_lyr = in_lyr

    def input_srs(self) -> SpatialReference:
        if self._in_srid is None:
            in_spatial_ref = self._srs_from_layer(self._in_lyr, self._ds_name)
        else:
            in_spatial_ref = osr.SpatialReference()
            import_res = in_spatial_ref.ImportFromEPSG(self._in_srid)
            if import_res != 0:
                raise SRIDNotFound(f'ESPG code {self._in_srid} not found in EPSG support files. This EPSG is invalid.')
        return in_spatial_ref

    def output_srs(self) -> SpatialReference:
        out_spatial_ref = osr.SpatialReference()
        out_spatial_ref.ImportFromEPSG(self._out_srid)
        return out_spatial_ref


#!-*-coding:utf-8-*-

import os
import json
from typing import Generator
from uuid import uuid4
from osgeo import ogr, osr
from osgeo.ogr import DataSource, Feature
from vectorio.vector.interfaces.ivector_file import IVectorFile
from vectorio.vector._src.generators.feature_collection_concatenated import (
    FeatureCollectionConcatenated
)
from vectorio.vector.exceptions import (
    ShapefileInvalid, ShapefileIsEmpty,
    ImpossibleCreateShapefileFromGeometryCollection, FileNotFound
)
from vectorio.vector.shapefile.file_required_by_extension import (
    FileRequiredByExtension
)
from vectorio.vector._src.gdal_aux.cloned_ds import (
    GDALClonedDataSource
)
from vectorio.vector.shapefile.encodings import ShapeEncodings
from vectorio.config import GDAL_DRIVERS_NAME


class Shapefile(IVectorFile):

    _driver = None
    _shape_encoding = None
    _search_encoding = None

    def __init__(self, search_encoding=True, search_encoding_exception=True):
        self._driver = ogr.GetDriverByName(GDAL_DRIVERS_NAME['ESRI Shapefile'])
        self._search_encoding = search_encoding
        self._shape_encoding = ShapeEncodings(
            raise_exception=search_encoding_exception
        )

    def _has_data(self, ds: DataSource):
        lyr = ds.GetLayer()
        if lyr.GetFeature(0) is None:
            raise ShapefileIsEmpty(
                "Shapefile is empty. Please, check if your shapefile has data."
            )

    def datasource(self, fpath: str) -> DataSource:
        if not os.path.exists(fpath):
            raise FileNotFound(f'"{fpath}" does not exists.')

        if self._search_encoding:
            os.environ['SHAPE_ENCODING'] = self._shape_encoding.from_file(
                fpath.replace('.shp', '.dbf')  # getting path from .dbf
            )

        ds = self._driver.Open(fpath)

        if ds is None:
            raise ShapefileInvalid(
                'Shapefile invalid. Please, check if your shapefile is correct'
                ' or if the files .dbf .shx and .prj are next to the .shp file.'
            )
        self._has_data(ds)
        return GDALClonedDataSource(ds).ref()

    def _export_json(self, feat: Feature) -> str:
        return json.dumps(
            json.loads(feat.ExportToJson()), ensure_ascii=False
        )

    def items(self, datasource: DataSource) -> Generator[str, None, None]:
        lyr = datasource.GetLayer(0)
        for idx_feat in range(lyr.GetFeatureCount()):
            feat = lyr.GetFeature(idx_feat)
            yield self._export_json(feat)

    def collection(self, datasource: DataSource) -> str:
        return FeatureCollectionConcatenated(self.items(datasource))

    def _create_prj(self, out_prj: str, feat: Feature):
        srs = feat.geometry().GetSpatialReference()
        with open(out_prj, 'w') as f:
            f.write(srs.ExportToWkt())

    def write(self, ds: DataSource, out_path: str,) -> str:
        assert out_path.endswith('.shp'), 'Output file have has .shp extension.'
        lyr = ds.GetLayer()
        feat = lyr.GetFeature(0)

        if feat.geometry().GetGeometryName() == 'GEOMETRYCOLLECTION':
            raise ImpossibleCreateShapefileFromGeometryCollection(
                'Impossible create shapefile from a geometry collection.'
                ' Please, convert the geometry collection for feature '
                'collection with same geometry type.'
            )
        ds_out = self._driver.CreateDataSource(out_path)
        ds_out.CopyLayer(ds.GetLayer(), str(uuid4()))

        out_prj = out_path.replace('.shp', '.prj')
        if not os.path.exists(out_prj):
           self._create_prj(out_prj, feat)

        ds_out.Destroy()
        return out_path



#!-*-coding:utf-8-*-

from uuid import uuid4
from typing import Generator
from functools import reduce

from osgeo import ogr, osr
from osgeo.ogr import Geometry, DataSource, Feature

from vectorio.vector.interfaces.ivector_data import IVectorData
from vectorio.vector.exceptions import WKTInvalid
from vectorio.vector.wkt.geom_type_factory import GeometryTypeFactory
from vectorio.config import GDAL_DRIVERS_NAME


GEOMETRYCOLLECTION_PREFIX = 'GEOMETRYCOLLECTION'


class WKT(IVectorData):

    _gt_factory = None
    _initial_srid = 0
    _as_geometry_collection = True

    def __init__(self, as_geometry_collection=True, srid=4326):
        self._gt_factory = GeometryTypeFactory()
        self._initial_srid = srid
        self._as_geometry_collection = as_geometry_collection

    def _srs(self):
        srs = osr.SpatialReference()
        srs.ImportFromEPSG(self._initial_srid)
        return srs

    def datasource(self, input_data: str) -> DataSource:
        drv = ogr.GetDriverByName(GDAL_DRIVERS_NAME['MEMORY'])
        out_ds = drv.CreateDataSource(str(uuid4()))
        geom = None
        if not bool(input_data):
            raise WKTInvalid(
                'Invalid wkt data. Please, check if data is a wkt valid.'
            )

        geom = ogr.CreateGeometryFromWkt(input_data)

        if geom is None:
            raise WKTInvalid(
                f'Invalid wkt data. Please, check is the data "{input_data}" is in wkt pattern.'
            )
        l1 = out_ds.CreateLayer(
            str(uuid4()), self._srs(),
            self._gt_factory.get_type(geom.GetGeometryName())
        )
        feat = ogr.Feature(l1.GetLayerDefn())
        feat.SetGeometry(geom)
        # this change on this layer "l1" will be reflected on "_out_ds"
        l1.SetFeature(feat)
        return out_ds

    def items(self, ds: DataSource) -> Generator[str, None, None]:
        lyr = ds.GetLayer(0)
        if lyr.GetFeatureCount() == 0:
            yield 'GEOMETRY_EMPTY'

        for idx_feat in range(lyr.GetFeatureCount()):
            feat = lyr.GetFeature(idx_feat)
            geom = feat.geometry()
            if geom is not None:
                yield geom.ExportToWkt()

    def collection(self, ds: DataSource) -> str:
        out_wkt = reduce(
            lambda x,y: x + ',' + y, self.items(ds)
        )
        if out_wkt == 'GEOMETRY_EMPTY':
            return 'GEOMETRY_EMPTY'

        if out_wkt.startswith(GEOMETRYCOLLECTION_PREFIX):
            return out_wkt
        else:
            if self._as_geometry_collection:
                return f'{GEOMETRYCOLLECTION_PREFIX} ({out_wkt})'
            else:
                return out_wkt

    def write(self, ds: DataSource, out_path: str) -> str:
        self._validate_basedir(out_path)
        return self._write_by_collection(ds, out_path)



#!-*-coding:utf-8-*-

from uuid import uuid4
from osgeo.ogr import DataSource


class GDALClonedLayer:

    _ds = None
    _another_layer = None

    def __init__(self, ds: DataSource, another_layer):
        self._ds = ds
        self._another_layer = another_layer

    def ref(self):
        another_layer_defn = self._another_layer.GetLayerDefn()
        out_lyr = self._ds.CreateLayer(
            str(uuid4()), self._another_layer.GetSpatialRef(),
            geom_type=self._another_layer.GetGeomType()
        )
        for i in range(another_layer_defn.GetFieldCount()):
            out_lyr.CreateField(another_layer_defn.GetFieldDefn(i))
        return out_lyr



from vectorio.vector.datasource.ds_reprojected import DataSourceReprojected
from vectorio.vector.shapefile.shapefile import Shapefile
from vectorio.vector.shapefile.decorators.shapefile_as_zip import ShapefileAsZip
from vectorio.vector.shapefile.decorators.shapefile_as_rar import ShapefileAsRar
from vectorio.vector.wkt.wkt import WKT
from vectorio.vector.geofile.geo_file import GeoFile
from vectorio.vector.geojson.geojson import Geojson
from vectorio.vector.composite.vector_composite import VectorComposite
from vectorio.vector.srs.reprojection import VectorReprojected
from vectorio.vector.utm.utm_zone import UTMZone



#!-*-coding:utf-8-*-

from osgeo import ogr


class GDALClonedFeature:

    _gdal_feature_ref = None
    _geom = None
    _layer_dfn = None
    _another_feature = None

    def __init__(self, layer_dfn, geom, another_feature):
        self._geom = geom
        self._layer_dfn = layer_dfn
        self._another_feature = another_feature

    def ref(self):
        gdal_feature_ref = ogr.Feature(self._layer_dfn)
        gdal_feature_ref.SetGeometry(self._geom)

        for i in range(self._layer_dfn.GetFieldCount()):
            field_name = self._layer_dfn.GetFieldDefn(i).GetNameRef()
            field_value = self._another_feature.GetField(i)
            gdal_feature_ref.SetField(field_name, field_value)

        return gdal_feature_ref



#!-*-coding:utf-8-*-

from abc import abstractmethod
from osgeo.ogr import DataSource
from typing import Generator
from vectorio.vector.interfaces.ivector import IVector


class IVectorData(IVector):

    @abstractmethod
    def datasource(self, input_data: str) -> DataSource:
        pass

    @abstractmethod
    def items(self, datasource: DataSource) -> Generator[str, None, None]:
        pass

    @abstractmethod
    def collection(self, datasource: DataSource) -> str:
        pass

    @abstractmethod
    def write(self, ds: DataSource, out_path: str) -> str:
        pass

    def _write_by_collection(self, ds: DataSource, out_path: str) -> str:
        with open(out_path, 'w') as f:
            f.write(self.collection(ds))
        return out_path



#!-*-coding:utf-8-*-
import os
from osgeo import ogr
from osgeo.ogr import DataSource
from vectorio.vector import (
    Shapefile, ShapefileAsZip, WKT
)
from vectorio.config import STATIC_DIR
from vectorio.vector.exceptions import (
    ErrorOnIntersection, ExistsManyGeometriesTypes, DataSourceNotIntersectsWithAnyUTMZones
)
from vectorio.config import GDAL_DRIVERS_NAME, PRJ_WGS84
from vectorio.vector import DataSourceReprojected


class UTMZone:

    _drv_mem = None
    _world_utm_grid_ds = None

    def __init__(self):
        os.environ['SHAPE_ENCODING'] = 'ISO-8859-3'
        self._world_utm_grid_ds = ShapefileAsZip(
            Shapefile(search_encoding=False)
        ).datasource(
            os.path.join(STATIC_DIR, 'World_UTM_Grid_HM.zip')
        )
        self._drv_mem = ogr.GetDriverByName(GDAL_DRIVERS_NAME['MEMORY'])

    def _intersection_ds(self, ds_utm: DataSource, inp_ds: DataSource):
        lyr_utm = ds_utm.GetLayer(0)
        target_lyr = inp_ds.GetLayer(0)
        ds_out = self._drv_mem.CreateDataSource('out_mem_ds')
        lyr_out = ds_out.CreateLayer('out_mem_Lyr')
        intersection = lyr_utm.Intersection(target_lyr, lyr_out)

        if intersection != 0:
            raise ErrorOnIntersection(
                "Error on Intersection. Please, check if your datasource is valid."
            )
        return ds_out

    def zones(self, inp_ds: DataSource) -> set:
        ds = self._intersection_ds(self._world_utm_grid_ds, inp_ds)
        lyr = ds.GetLayer(0)
        result = set()

        for idx_feat in range(lyr.GetFeatureCount()):
            feat = lyr.GetFeature(idx_feat)
            zone = f'{feat.GetFieldAsInteger("ZONE")}{feat.GetFieldAsString("HEMISPHERE")}'
            result.add(zone)
        return result

    def zone_from_biggest_geom(self, inp_ds: DataSource, wkt_prj_for_metrics: str, in_wkt_prj=PRJ_WGS84):
        inter_ds = self._intersection_ds(self._world_utm_grid_ds, inp_ds)
        inter_lyr = inter_ds.GetLayer()

        if inter_lyr.GetFeatureCount() == 0:
            raise DataSourceNotIntersectsWithAnyUTMZones(
                'The datasource not intersect with UTM Zones. Please, check '
                'if "in_wkt_prj" is equivalent to Datum from datasource. Use '
                'Datuns for this operations.'
            )

        ds = DataSourceReprojected(
            inter_ds, in_wkt_prj=in_wkt_prj,
            out_wkt_prj=wkt_prj_for_metrics, use_wkt_prj=True
        ).ref()
        biggest_part_idx = 0
        indexes_from_feats = []  # all indexes from features
        metrics = []  # area or length of the geometry from feature
        lyr = ds.GetLayer(0)

        for idx_feat in range(lyr.GetFeatureCount()):
            feat = lyr.GetFeature(idx_feat)
            geom_type = feat.geometry().GetGeometryName()
            area_or_length = 0
        
            if geom_type == 'POLYGON':
                area_or_length = feat.geometry().Area()
            elif geom_type == 'LINESTRING':
                area_or_length = feat.geometry().Length()
            else:
                area_or_length = 0

            indexes_from_feats.append(idx_feat)
            metrics.append(area_or_length)

        # getting index from biggest geometry by feature
        biggest_part_idx = indexes_from_feats[
            metrics.index(  # getting index by value
                max(metrics)  # getting max area or length
            )
        ]
        feat = lyr.GetFeature(biggest_part_idx) # big geometry
        zone = f'{feat.GetFieldAsInteger("ZONE")}{feat.GetFieldAsString("HEMISPHERE")}'
        return zone



#!-*-coding:utf-8-*-

from abc import ABC, abstractmethod
from osgeo.ogr import DataSource
from typing import Generator
from vectorio.vector.interfaces.ivector import IVector


class IVectorFile(IVector):

    @abstractmethod
    def datasource(self, fpath: str) -> DataSource:
        pass

    @abstractmethod
    def items(self, datasource: DataSource) -> Generator[str, None, None]:
        pass

    @abstractmethod
    def collection(self, datasource: DataSource) -> str:
        pass

    @abstractmethod
    def write(self, ds: DataSource, out_path: str) -> str:
        pass



#!-*-coding:utf-8-*-

import os


def get_root_project(directory):
    path = os.path.abspath(directory)
    if os.path.isdir(path):
        if "vectorio" in os.listdir(path):
            return path
        else:
            return get_root_project(os.path.dirname(path))
    else:
        return get_root_project(os.path.dirname(path))


ROOT_PROJECT = get_root_project(__file__)
STATIC_DIR = os.path.join(ROOT_PROJECT, 'vectorio', '_assets')
GDAL_DRIVERS_NAME = {
    'MEMORY': 'MEMORY',
    'ESRI Shapefile': 'ESRI Shapefile',
    'GeoJSON': 'GeoJSON'
}

PRJ_WGS84='GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.01745329251994328,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]]'


#!-*-coding:utf-8-*-

from vectorio.vector.composite.interfaces.ivector_composite import (
    IVectorComposite
)
from vectorio.vector.interfaces.ivector import IVector
from osgeo.ogr import DataSource
from typing import Generator


class VectorComposite(IVectorComposite):

    def __init__(self, input_vector: IVector, output_vector: IVector):
        self._input_vector = input_vector
        self._output_vector = output_vector

    def items(self, input_data: str) -> Generator[str, None, None]:
        return self._output_vector.items(
            self._input_vector.datasource(input_data)
        )

    def collection(self, input_data: str) -> str:
        return self._output_vector.collection(
            self._input_vector.datasource(input_data)
        )

    def write(self, input_data: str, out_path: str) -> str:
        return self._output_vector.write(
            self._input_vector.datasource(input_data), out_path
        )



#!-*-coding:utf-8-*-

import os
import copy

from vectorio.vector.exceptions import FileNotFound
from vectorio.vector.shapefile.interfaces.ifile_required_by_extension import (
    IFileRequiredByExtension
)


class FileRequiredByExtension(IFileRequiredByExtension):

    def __init__(self, dirpath: str, expected_ext_list: list):
        self._dirpath = dirpath
        self._expected_ext_list = expected_ext_list

    def _validate_found_ext(self, found_ext: list):
        """
        This method will check if the required files are found. Case the some
        file is missing, a exception will be raised.
        """
        missing_files = list(set(self._expected_ext_list) - set(found_ext))
        if len(missing_files) != 0:
            raise FileNotFound(f"File .{''.join(missing_files)} not found.")

    def _files_by_ext(self) -> dict:
        """
        This method will iterate by directory (dirpath) for search (recursivly)
        the files by extension.
        """
        def find_recursivly(dirpath, ext_lst):
            found_files_by_ext = {}  # result of this method
            missing_files = copy.copy(ext_lst)  # list of extension from shapefile

            for fname in os.listdir(dirpath):
                fpath = os.path.join(dirpath, fname)

                if len(missing_files) == 0:
                    return found_files_by_ext

                if os.path.isdir(fpath):
                    return find_recursivly(fpath, missing_files)
                else:
                    ext = fname.split('.')[-1].lower()
                    if ext in missing_files:  # found extension
                        missing_files.pop(missing_files.index(ext))
                        found_files_by_ext[ext] = fpath
            return found_files_by_ext

        return find_recursivly(self._dirpath, self._expected_ext_list)

    def files(self):
        files = self._files_by_ext()
        self._validate_found_ext(files.keys())
        return files

    def __repr__(self):
        return '<FileRequiredByExtension>'



#!-*-coding:utf-8-*-

import os
import shutil
from typing import Generator
from osgeo.ogr import DataSource
from vectorio.vector.interfaces.ivector_file import IVectorFile
from zipfile import ZipFile
from vectorio.vector.shapefile.file_required_by_extension import (
    FileRequiredByExtension
)
import tempfile


class ShapefileAsZip(IVectorFile):

    _shapefile = None

    def __init__(self, shapefile: IVectorFile):
        self._shapefile = shapefile

    def _extraction_dir(self, fpath: str) -> str:
        tmpdir = tempfile.mkdtemp()
        with ZipFile(fpath) as zipf:
            zipf.extractall(tmpdir)
        return tmpdir

    def _compress_files(self, fpath: str, files: list) -> str:
        with ZipFile(fpath, 'w') as zipf:
            for f in files:
                zipf.write(f)
        return fpath

    def datasource(self, fpath: str) -> DataSource:
        dir_with_files = self._extraction_dir(fpath)
        files_required = FileRequiredByExtension(
            dir_with_files, ['shp', 'dbf', 'shx', 'prj']
        )
        ds = self._shapefile.datasource(files_required.files()['shp'])
        shutil.rmtree(dir_with_files)
        return ds

    def items(self, datasource: DataSource) -> Generator[str, None, None]:
        return self._shapefile.items(datasource)

    def collection(self, datasource: DataSource):
        return self._shapefile.collection(datasource)

    def write(self, ds: DataSource, out_path: str) -> str:
        assert out_path.endswith('.zip'), 'Output file have has .zip extension.'
        out_shp = self._shapefile.write(ds, out_path.replace('.zip', '.shp'))
        files = [
            out_shp,
            out_shp.replace('.shp', '.dbf'),
            out_shp.replace('.shp', '.prj'),
            out_shp.replace('.shp', '.shx')
        ]
        out_zip = self._compress_files(out_path, files)
        for f in files:
            os.remove(f)
        return out_zip



#!-*-coding:utf-8-*-

from uuid import uuid4
from osgeo import ogr, osr
from osgeo.ogr import DataSource, Layer
from osgeo.osr import SpatialReference
from vectorio.vector._src.gdal_aux.cloned_feature import GDALClonedFeature
from vectorio.vector._src.gdal_aux.cloned_layer import GDALClonedLayer
from vectorio.vector.exceptions import DataSourceWithoutSpatialRef
from vectorio.config import GDAL_DRIVERS_NAME
from vectorio.vector.srs.srs import SRS


class DataSourceReprojected:

    _inp_ds = None
    _in_srid = None
    _out_srid = None
    _in_wkt_prj = None
    _out_wkt_prj = None
    _use_wkt_prj = None

    def __init__(
        self, inp_ds: DataSource, in_srid=None, out_srid=None,
        in_wkt_prj=None, out_wkt_prj=None, use_wkt_prj=False
    ):
        self._inp_ds = inp_ds
        self._in_srid = in_srid
        self._out_srid = out_srid
        self._in_wkt_prj = in_wkt_prj
        self._out_wkt_prj = out_wkt_prj
        self._use_wkt_prj = use_wkt_prj
        self._driver_mem = ogr.GetDriverByName(GDAL_DRIVERS_NAME['MEMORY'])

    def ref(self):
        inp_lyr = self._inp_ds.GetLayer()
        srs = SRS(
            self._inp_ds.name, inp_lyr, self._in_srid, self._out_srid,
            self._in_wkt_prj, self._out_wkt_prj, self._use_wkt_prj
        )

        coord_transformator = srs.coord_transformator()
        out_ds = self._driver_mem.CreateDataSource(str(uuid4()))
        out_lyr = GDALClonedLayer(out_ds, inp_lyr).ref()
        outLayerDefn = out_lyr.GetLayerDefn()
        in_feature = inp_lyr.GetNextFeature()

        while in_feature:
            geom = in_feature.GetGeometryRef()
            if geom is not None:
                geom.Transform(coord_transformator)
                gdal_feature = GDALClonedFeature(outLayerDefn, geom, in_feature)
                out_lyr.CreateFeature(gdal_feature.ref())
            in_feature = inp_lyr.GetNextFeature()

        return out_ds



#!-*-coding:utf-8-*-

from osgeo import gdal


class GDALCustomError:

    def __init__(self):
        self.level = gdal.CE_None
        self.code = 0
        self.msg = ''

    def handler(self, err_level: str, err_no: int, err_msg: str):
        self.level=err_level
        self.code=err_no
        self.msg=err_msg




#!-*-coding:utf-8-*-
import os
from abc import ABC
from vectorio.vector.metrics.vertices import Vertices


class IVector(ABC):

    def vertices_by_feature(self, ds):
        vertices = Vertices()
        lyr = ds.GetLayer(0)
        vertices_count = {}
        for feat_idx in range(lyr.GetFeatureCount()):
            feat = lyr.GetFeature(feat_idx)
            geom = feat.geometry()
            feat_id = f'Feature_{feat_idx}_{geom.GetGeometryName()}'
            vertices_count[feat_id] = vertices.tell(geom)
        return vertices_count

    def _validate_basedir(self, out_path: str):
        dirname = os.path.dirname(out_path)
        if dirname != '':
            if not os.path.exists(dirname):
                raise DirectoryDoesNotExists(
                    f'The directory "{dirname}" does not exists.'
                )



#!-*-coding:utf-8-*-

from osgeo import ogr
from osgeo.ogr import DataSource
from vectorio.config import GDAL_DRIVERS_NAME

class GDALClonedDataSource:

    _inp_ds = None
    _driver_mem = None
    _LYR_NAME = 'Cloned Layer 01'

    def __init__(self, inp_ds: DataSource):
        self._driver_mem = ogr.GetDriverByName(GDAL_DRIVERS_NAME['MEMORY'])
        self._inp_ds = inp_ds

    def ref(self):
        inp_lyr = self._inp_ds.GetLayer()
        out_ds = self._driver_mem.CreateDataSource('out')
        out_ds.CopyLayer(inp_lyr, self._LYR_NAME)
        return out_ds





class Vertices:

    _FUNC_FOR_COUNT_VERTICES = {}

    def __init__(self):
        # "switch" method for count vertices
        self._FUNC_FOR_COUNT_VERTICES = {
            'POINT': self._point,
            'LINESTRING': self._line,
            'POLYGON': self._polygon,
            'MULTIPOINT': self._multipoint,
            'MULTILINESTRING': self._multiline,
            'MULTIPOLYGON': self._multipolygon
        }

    def _multiline(self, geom):
        vertices = []
        for idx_geom in range(geom.GetGeometryCount()):
            vertices.append(geom.GetGeometryRef(idx_geom).GetPointCount())
        return vertices

    def _multipoint(self, geom):
        vertices = []
        for idx_geom in range(geom.GetGeometryCount()):
            vertices.append(geom.GetGeometryRef(idx_geom).GetPointCount())
        return vertices

    def _multipolygon(self, geom):
        vertices = []
        rings = []
        for idx_geom in range(geom.GetGeometryCount()):
            polygon = geom.GetGeometryRef(idx_geom)
            rings = []
            for idx_ring in range(polygon.GetGeometryCount()):
                ring = polygon.GetGeometryRef(idx_ring)
                #print(ring.GetPointCount())
                rings.append(ring.GetPointCount())
            vertices.append(rings)
        return vertices

    def _point(self, geom):
        vertices = []
        vertices.append(geom.GetPointCount())
        return vertices

    def _line(self, geom):
        vertices = []
        vertices.append(geom.GetPointCount())
        return vertices

    def _polygon(self, geom):
        vertices = []
        for idx_geom in range(geom.GetGeometryCount()):
            vertices.append(geom.GetGeometryRef(idx_geom).GetPointCount())
        return vertices

    def tell(self, geom):
        vertices_counter = self._FUNC_FOR_COUNT_VERTICES.get(geom.GetGeometryName())
        if vertices_counter is None:
            raise Exception('')
        return vertices_counter(geom)



#!-*-coding:utf-8-*-

from osgeo.ogr import DataSource
from typing import Generator
from vectorio.vector.interfaces.ivector_data import IVectorData
from vectorio.vector.interfaces.ivector_file import IVectorFile


class GeoFile(IVectorFile):

    _vector = None

    def __init__(self, vector: IVectorData):
        self._vector = vector

    def datasource(self, fpath: str) -> DataSource:
        with open(fpath) as geof:
            return self._vector.datasource(geof.read())

    def items(self, datasource: DataSource) -> Generator[str, None, None]:
        return self._vector.items(datasource)

    def collection(self, datasource: DataSource) -> str:
        return self._vector.collection(datasource)

    def write(self, ds: DataSource, out_path: str) -> str:
        self._vector.write(ds, out_path)



#!-*-coding:utf-8-*-

from functools import reduce


class FeatureCollectionConcatenated(str):

    def __new__(cls, *args, **kwargs):
        features_gen = args[0]
        concat_features_lbd = lambda x, y: str(x) + ',' + str(y)
        features_str = reduce(concat_features_lbd, features_gen)
        return '{"type": "FeatureCollection","features": [' + features_str + ']}'


#!-*-coding:utf-8-*-

import os
import warnings
from vectorio.vector.exceptions import CharDecodeError


SHAPEFILE_ENCODINGS = {
    'LATIN1': 'ISO-8859-1', 'LATIN3': 'ISO-8859-3', 'UTF-8': 'UTF-8',
    'ISO-8859-2': 'ISO-8859-2'
}
ANY_INDEX = 4


class ShapeEncodings:

    _raise_exception = None

    def __init__(self, raise_exception=True):
        self._raise_exception = raise_exception

    def _accepted_shape_encodings(self):
        return str(
            list(
                SHAPEFILE_ENCODINGS.values()
            )
        ).replace("'", "").replace("[", "").replace("]", "")

    def _validate_encoding(self, fpath_dbf: str, encoding: str):
        with open(fpath_dbf, encoding=encoding) as fdbf:
            fdbf.read()[0:ANY_INDEX]

    def from_file(self, fpath_dbf: str) -> str:
        encoding = SHAPEFILE_ENCODINGS['UTF-8']
        try:
            self._validate_encoding(fpath_dbf, encoding)
        except UnicodeDecodeError:
            encoding = SHAPEFILE_ENCODINGS['LATIN1']
            self._validate_encoding(fpath_dbf, encoding)
        except UnicodeDecodeError:
            encoding = SHAPEFILE_ENCODINGS['LATIN3']
            self._validate_encoding(fpath_dbf, encoding)
        except UnicodeDecodeError:
            encoding = SHAPEFILE_ENCODINGS['ISO-8859-2']
            self._validate_encoding(fpath_dbf, encoding)
        except Exception as ex:
            MSG = f"Erro on read {fpath_dbf}. Please, check if the " \
                f"{fpath_dbf} file exists and check the encoding " \
                "character of the .dbf.\n" \
                f"Accepted encodings: {self._accepted_shape_encodings()}."\
                f"Inner Exception: {str(ex)}"

            if self._raise_exception:
                raise CharDecodeError(MSG)

        return encoding



#!-*-coding:utf-8-*-

from typing import Generator
from osgeo.ogr import DataSource
from osgeo import ogr, osr
from vectorio.vector.interfaces.ivector import IVector
from vectorio.vector import DataSourceReprojected


class VectorReprojected(IVector):

    _vector = None
    _in_srid = None
    _out_srid = None

    def __init__(
        self, vector: IVector, in_srid: int=None, out_srid: int=None
    ):
        self._vector = vector
        self._in_srid = in_srid
        self._out_srid = out_srid

    def datasource(self, input_data: str) -> DataSource:
        return DataSourceReprojected(
            self._vector.datasource(input_data),
            self._in_srid, self._out_srid
        ).ref()

    def items(self, datasource: DataSource) -> Generator[str, None, None]:
        return self._vector.items(datasource)

    def collection(self, datasource: DataSource) -> str:
        return self._vector.collection(datasource)

    def write(self, ds: DataSource, out_path: str) -> str:
        return self._vector.write(ds, out_path)



#!-*-coding:utf-8-*-

from osgeo import ogr
from vectorio.vector.wkt.interfaces.igeom_type_factory import (
    IGeometryTypeFactory
)
from vectorio.vector.exceptions import WKTInvalid, GeometryTypeInvalid


DATA = {
    'POINT': ogr.wkbPoint,
    'LINESTRING': ogr.wkbLineString,
    'POLYGON': ogr.wkbPolygon,
    'MULTIPOINT': ogr.wkbMultiPoint,
    'MULTILINESTRING': ogr.wkbMultiPoint,
    'MULTIPOINT': ogr.wkbMultiPoint,
    'MULTILINESTRING': ogr.wkbMultiLineString,
    'MULTIPOLYGON': ogr.wkbMultiPolygon,
    'GEOMETRYCOLLECTION': ogr.wkbGeometryCollection,
}


class GeometryTypeFactory(IGeometryTypeFactory):

    def get_type(self, input_geom_type):
        geom_type = DATA.get(input_geom_type)
        if geom_type is None:
            raise GeometryTypeInvalid(
                f'The geometry type "{input_geom_type}" isn\'t registered in this project'
            )
        return geom_type



from setuptools import setup, find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(name='vectorio',
    version='1.2.5',
    description='Geoprocessing utility for working with vector data',
    author='Igor Rodrigues Sousa Silva',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author_email='igor.rodrigues.ss10@gmail.com',
    url='https://github.com/igor-rodrigues-ss/vectorio',
    test_suite='tests',
    packages=find_packages(exclude='tests'),
    tests_require=['pytest'],
    install_requires=[
        'rarfile==3.1',
    ],
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation ",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
    ],
    python_requires='>=3.6',
    zip_safe=False,
)


#!-*-coding:utf-8-*-

import os
import json
from typing import Generator
from osgeo.ogr import DataSource, Feature
from osgeo import ogr
from vectorio.vector.interfaces.ivector_data import IVectorData
from vectorio.vector.exceptions import GeojsonInvalid
from vectorio.vector._src.generators.feature_collection_concatenated import (
    FeatureCollectionConcatenated
)
from vectorio.config import GDAL_DRIVERS_NAME

class Geojson(IVectorData):

    _driver = None

    def __init__(self):
        self._driver = ogr.GetDriverByName(GDAL_DRIVERS_NAME['GeoJSON'])

    def datasource(self, input_data: str) -> DataSource:
        ds = self._driver.Open(input_data)
        if ds is None:
            raise GeojsonInvalid(
                'Invalid geojson data. Plese check if the data is on geojson pattern.'
            )
        return ds

    def items(self, ds: DataSource) -> Generator[str, None, None]:
        lyr = ds.GetLayer(0)
        for idx_feat in range(lyr.GetFeatureCount()):
            feat = lyr.GetFeature(idx_feat)
            yield json.dumps(
                json.loads(feat.ExportToJson()), ensure_ascii=False
            )

    def collection(self, ds: DataSource) -> str:
        return FeatureCollectionConcatenated(self.items(ds))

    def write(self, ds: DataSource, out_path: str) -> str:
        self._validate_basedir(out_path)
        return self._write_by_collection(ds, out_path)



#!-*-coding:utf-8-*-

from vectorio.vector.srs.srid import SRSFromSRID
from vectorio.vector.srs.wkt_prj import SRSFromWKTPRJ
from osgeo.ogr import Layer
from osgeo import osr


class SRS:

    _in_srid = None
    _out_srid = None
    _in_wkt_prj = None
    _out_wkt_prj = None
    _use_wkt_prj = None
    _in_lyr = None

    def __init__(
        self, ds_name: str, in_lyr: Layer, in_srid: int=None, out_srid: int=None,
        in_wkt_prj: str=None, out_wkt_prj: str=None, use_wkt_prj=False
    ):
        self._in_srid = in_srid
        self._out_srid = out_srid
        self._in_wkt_prj = in_wkt_prj
        self._out_wkt_prj = out_wkt_prj
        self._use_wkt_prj = use_wkt_prj
        self._ds_name = ds_name
        self._in_lyr = in_lyr

    def coord_transformator(self):
        srs = None

        if self._use_wkt_prj:
            assert self._out_wkt_prj != None, 'Is required output WKT projection'
            srs = SRSFromWKTPRJ(
                self._ds_name, self._in_lyr, in_wkt_prj=self._in_wkt_prj, out_wkt_prj=self._out_wkt_prj
            )
        else:
            assert self._out_srid != None, 'Is required output SRID'
            srs = SRSFromSRID(
                self._ds_name, self._in_lyr, in_srid=self._in_srid, out_srid=self._out_srid
            )    
        in_srs = srs.input_srs()
        out_srs = srs.output_srs()
        return osr.CoordinateTransformation(in_srs, out_srs)